/*-------------------------------------------------------------------------
 *
 * lwlocktest.c
 *
 * Copyright (c) 2012, PostgreSQL Global Development Group
 *
 * IDENTIFICATION
 *	  contrib/lwlocktest/lwlocktest.c
 *
 *-------------------------------------------------------------------------
 */

#include "postgres.h"

#include "storage/barrier.h"
#include "storage/lwlock.h"
#include "funcapi.h"
#include "miscadmin.h"

PG_MODULE_MAGIC;

Datum		lwlocktest(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(lwlocktest);

typedef struct
{
	LWLock *lock1;
	LWLock *lock2;
	uint64 lock1value;
} lwlocktest_shared;

static lwlocktest_shared *shared;
static LWLock *lock1;
static LWLock *lock2;

Datum
lwlocktest(PG_FUNCTION_ARGS)
{
	uint32		n = PG_GETARG_UINT32(0);
	bool		found;
	int			i;

	if (!shared)
	{
		LWLockAcquire(AddinShmemInitLock, LW_EXCLUSIVE);
		shared = (lwlocktest_shared *) ShmemInitStruct("lwlocktest", sizeof(lwlocktest_shared), &found);
		if (!found)
		{
			shared->lock1 = LWLockAssign();
			shared->lock2 = LWLockAssign();
			shared->lock1value = 0;
		}
		lock1 = shared->lock1;
		lock2 = shared->lock2;
		LWLockRelease(AddinShmemInitLock);
	}

	for (i = 0; i < 100000000; i++)
	{
		if (n == 1)
		{
			LWLockAcquireWithVar(lock1,
								 &shared->lock1value,
								 UINT64CONST(0xFFFFFFFFFFFFFFFF));
			LWLockAcquire(lock2, LW_EXCLUSIVE);
			LWLockRelease(lock1);
			LWLockRelease(lock2);
		}
		else if (n == 2)
		{
			uint64 newval;
			LWLockAcquire(lock2, LW_EXCLUSIVE);
			LWLockWaitForVar(lock1, &shared->lock1value, 0, &newval);
			LWLockRelease(lock2);
		}
		else
		{
			LWLockAcquireWithVar(lock1,
								 &shared->lock1value,
								 0);
			LWLockRelease(lock1);
		}
		if (i % 100000 == 0)
		{
			elog(NOTICE, "still looping %d (%d)...", n, i);
			CHECK_FOR_INTERRUPTS();
		}
	}

	PG_RETURN_VOID();
}
